// © 2023 John Breauxs
use super::{
    context::Context,
    token::{OwnedToken, *},
};
use std::fmt::Display;

#[derive(Debug)]
pub enum LexError {
    /// Any other error, tagged with [Context]. Created by [`Error::context()`]
    Contextual(Context, Box<Self>),
    /// Produced by [Token] when the input is entirely unexpected.
    UnexpectedSymbol(String),
    /// Produced by [`TokenStream::expect`] when the next [Token] isn't the expected [Type]
    UnexpectedToken { expected: Type, got: OwnedToken },
    /// Produced by [`TokenStream::expect_any_of`] when the next [Token] isn't any of the
    /// expected [Types](Type)
    AllExpectationsFailed { expected: Types, got: OwnedToken },
}

impl LexError {
    pub fn context(self, c: Context) -> Self {
        match self {
            Self::Contextual(..) => self,
            _ => Self::Contextual(c, Box::new(self)),
        }
    }

    // Extracts the root of the error tree
    pub fn bare(self) -> Self {
        match self {
            Self::Contextual(_, bare) => bare.bare(),
            _ => self,
        }
    }

    pub fn expected<E: AsRef<[Type]>, T: Into<OwnedToken>>(expected: E, got: T) -> Self {
        match expected.as_ref().len() {
            1 => Self::UnexpectedToken { expected: expected.as_ref()[0], got: got.into() },
            _ => Self::AllExpectationsFailed { expected: expected.as_ref().into(), got: got.into() },
        }
    }

    pub fn mask_expectation(mut self, expected: Type) -> Self {
        match self {
            LexError::UnexpectedToken { got, .. } => self = LexError::UnexpectedToken { expected, got },
            LexError::AllExpectationsFailed { got, .. } => self = LexError::UnexpectedToken { expected, got },
            LexError::Contextual(context, err) => {
                self = LexError::Contextual(context, Box::new(err.mask_expectation(expected)))
            }
            _ => (),
        }
        self
    }
}

impl Display for LexError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            LexError::Contextual(ctx, error) => write!(f, "{ctx}: {error}"),
            LexError::UnexpectedSymbol(sym) => write!(f, "Unexpected item in bagging area: \"{sym}\""),
            LexError::UnexpectedToken { expected, got } => write!(f, "Expected {expected}, got {got}."),
            LexError::AllExpectationsFailed { expected, got } => write!(f, "Expected {expected}, got {got}."),
        }
    }
}

impl std::error::Error for LexError {}
